const buffer = require('buffer');
const { TextDecoder } =  require('util');

let stream = [];

const inverseCppBytes = (cppBytes) => {
    let jsBytes = [];
    // invert each 2 bytes inside js array
    for(let i = 0; i < cppBytes.length; i+=2) {
        jsBytes.push(cppBytes[i+1]);
        jsBytes.push(cppBytes[i]);
    }
    return jsBytes;
}

const cppBytesToString = (cppBytes) => {
    let jsBytes = inverseCppBytes(cppBytes);
    const buf16 = Buffer.from(jsBytes, 'utf16le');
    return buffer.transcode(buf16, 'utf16le', 'utf8').toString();
}

const ReadByte = () => {
    if(stream.length < 1 )
        return '';
    const raw_bytes = stream.splice(0, 1);
    return raw_bytes[0];
};

const ReadChar = () => {
    if(stream.length < 1 )
        return '';
    const raw_bytes = stream.splice(0, 1);
    return raw_bytes[0];
};

const ReadBool = () => {
    if(stream.length < 1 )
        return false;
    const raw_bytes = stream.splice(0, 1);
    return raw_bytes[0] !== 0;
};

const ReadInt16 = () => {
    if(stream.length < 2 )
        return 0;
    const raw_bytes = stream.splice(0, 2);
    return ((raw_bytes[0] << 8) | raw_bytes[1]);
};

const ReadInt32 = () => {
    if(stream.length < 4 )
        return 0;
    const raw_bytes = stream.splice(0, 4);
    return ((raw_bytes[0] << 24) | (raw_bytes[1] << 16) | (raw_bytes[2] << 8) | raw_bytes[3]);
};

const ReadInt64 = () => {
    if(stream.length < 8 )
        return 0;
    const raw_bytes = stream.splice(0, 8);
    return ((raw_bytes[0] << 56) | (raw_bytes[1] << 48) | (raw_bytes[2] << 40) | (raw_bytes[3] << 32) |
        (raw_bytes[4] << 24) |  (raw_bytes[5] << 16) |(raw_bytes[6] << 8) | raw_bytes[7]);
};

const ReadString = () => {
    const size = ReadInt64();
    if(size === 0 )
        return '';
    const raw_bytes = stream.splice(0, 2 * size); // 2 symbols for unicode
    return cppBytesToString(raw_bytes);
};

const ReadWideString = () => {
    const size = ReadInt64();
    if(size === 0 )
        return '';
    const raw_bytes = stream.splice(0, 2 * size); // use 2 symbols for unicode character
    return cppBytesToString(raw_bytes);
};

const ReadBinary = () => {
    const size = ReadInt64();
    if(size === 0 )
        return { type:'binary', value: ''};
    const raw_bytes = stream.splice(0, size);
    const str = Buffer.from(raw_bytes).toString('base64');
    return { type:'binary', value: str };
};

const ReadArray = () => {
    let v = [];
    const count = ReadInt64();
    if(count === 0 )
        return v;
    for(let i = 0; i < count; i++){
        const type = ReadInt32();
        const item  = ReadTypedValue(type);
        if(item !== {} ){
            v.push(item);
        }
    }

    return v;
};

const ReadTypedValue = (type, settDT) => {
    if(type === 0 ) { // byte
        return ReadByte();
    } else if(type === 1 || type === 2) {            // typeidChar, typeidUnsignedChar
        return ReadChar();
    } else if(type === 3 || type === 4 ) {     // typeidInt16, typeidUnsignedInt16
        return ReadInt16();
    } else if(type === 5 || type === 6) {     // typeidInt32, typeidUnsignedInt32
        return ReadInt32();
    } else if(type === 7 || type === 8) {     // typeidInt64, typeidUnsignedInt64
        return ReadInt64();
    } else if(type === 9) {     // typeidString
        return ReadString();
    }  else if(type === 10) {     // typeidWString
        return ReadWideString();
    } else if(type === 11) {     // typeidBool
        return ReadBool();
    } else if(type === 12) {     // typeidBinary
        return ReadBinary();
    } else if(type === 13) {     // typeidArray
        return  ReadArray();
    } else if(type === 14) {     // typeidContainer
        return Deserialize(settDT);
    }
    return {};
}

const Deserialize = (settDT) => {
    let result = {};
    const tagCount = ReadInt64();
    for(let i = 0; i < tagCount; i++){
        const tag = ReadInt32();
        if(tag===0 )
            continue;

        const type = ReadInt32();
        const childDT = (settDT && settDT[tag]) ? settDT[tag].childs : undefined;
        const item  = ReadTypedValue(type, childDT);
        if(item !== {} ){
            let key = tag===65280 ? '__VersionInfo' : tag;
            if(settDT && settDT[tag] && settDT[tag].name) {
                key = settDT[tag].name;
            }
            result[key] = item;
        }
    }
    return result;
}

// exports
//
const deserialize = (bytes, settDT) => {
    stream = bytes;
    return Deserialize(settDT);
}

const blobToJson = (blob) => {
    try {
        const buf16 = Buffer.from(blob, 'base64');
        const decoded = buffer.transcode(buf16, 'utf16le', 'utf8').toString();
        return JSON.parse(decoded);
    }
    catch (e) {
    }
    return {};
}

module.exports = {
    deserialize,
    blobToJson
}