const hlpCrc = require('./utils_crc32.js');

const PlcControllerTypes = {
    1: 'Siemens Simatic S7-300',
    2: 'Siemens Simatic S7-400',
    8: 'Siemens Simatic S7-400H',
    3: 'Schneider Modicon M340',
    4: 'Schneider Modicon M340',
    5: 'Schneider Modicon M580',
    6: 'MitsubishiSystemQ',
    7: 'TestController',
};

const PlcPublisher = {
    1: 'Siemens',
    2: 'Siemens',
    8: 'Siemens',
    3: 'Schneider Electric',
    4: 'Schneider Electric',
    5: 'Schneider Electric',
    6: 'Mitsubishi Electric',
    7: 'TestController',
};

const siemens_types = [1,2,8];

const pad = (number, length) => {
    let str = '' + number;
    while (str.length < length) {
        str = '0' + str;
    }
    return str;
}

const GetPlcIdValue = (model) => {
    if(model===undefined)
        return 0;
    let out = model.ControllerType.toString() + model.IPAddress;
    out += pad(model.PortNumber, 5);
    if(siemens_types.includes(model.ControllerType)) {
        out += pad(model.RackNumber, 3);
        out += pad(model.SlotNumber, 3);
        out += model.ReadDBBlocks ? '1' : '0';
    } else {
        out += '0000000';
    }

    out += model.PasswordAvailable ? '1' : '0';
    if(model.PasswordAvailable){
        const buff = Buffer.from(model.OldPassword,'base64');
        const p_hash = hlpCrc.CalcBytesCrc32(buff);
        out += p_hash.toString();
    }

    return hlpCrc.CalcCrc32(out);
}

const GetLocalizedPlcType = (type) => {
    return PlcControllerTypes[type]!==undefined ? PlcControllerTypes[type] : '';
}

const GetLocalizedPlcPublisher = (type) => {
    return PlcPublisher[type]!==undefined ? PlcPublisher[type] : '';
}

const IsSiemensPlc = (type) => {
    return siemens_types.includes(type);
}

const GetPlcId = (model) => {
    if(model===undefined || model.ControllerType===undefined || model.ControllerType===0 )
        return '';
    const hash = GetPlcIdValue(model);
    return hash.toString(16).toUpperCase();
}

const hasRackSlotInput = (type) => { return (type === 1 || type === 2); }

const getPlcConnDescription = (model) =>{
    const localization = runtime.getSync('services.localization');
    let out = '';
    if(IsSiemensPlc(model.ControllerType)) {
        const block_usage = localization.get(model.ReadDBBlocks ? 'btns.Yes' : 'btns.No');
        out = localization.get('tasks.PlcInvestigate.fmts.full_conn',
            [model.IPAddress, model.PortNumber, model.RackNumber, model.SlotNumber, block_usage]);
    }
    else{
        out = localization.get('tasks.PlcInvestigate.fmts.base_conn', [model.IPAddress, model.PortNumber]);
    }

    if(model.PasswordAvailable){
        out += '; ';
        out += localization.get('tasks.PlcInvestigate.fmts.has_pwrd');
    }

    return out;
}

const DisplayedCheckConnectionArray = (cfg) => {
    let connects = [];
    if(cfg.ConnectionSettings.S7ConnectionsArray.length > 0 ){
        connects = cfg.ConnectionSettings.S7ConnectionsArray.map(function(item) {
            return {
                IPAddress: item.Base.IPAddress,
                PortNumber: item.Base.PortNumber,
                RackNumber: item.RackNumber.toString(),
                SlotNumber: item.SlotNumber.toString(),
            };
        });
    }
    else {
        connects = [{
            IPAddress: cfg.ConnectionSettings.S7Connection.Base.IPAddress,
            PortNumber: cfg.ConnectionSettings.S7Connection.Base.PortNumber,
            RackNumber: IsSiemensPlc(cfg.ControllerType) ? cfg.ConnectionSettings.S7Connection.RackNumber.toString() : '',
            SlotNumber: IsSiemensPlc(cfg.ControllerType) ? cfg.ConnectionSettings.S7Connection.SlotNumber.toString() : '',
        }];
    }

    return connects;
}

const toViewConnectionArray = (cfg) => {
    if(cfg.ControllerType !== 8) {
        return [];
    }
    // Siemens Simatic S7-400H
    let connects = [];
    if(cfg.ConnectionSettings.S7ConnectionsArray.length > 0 ){
        connects = cfg.ConnectionSettings.S7ConnectionsArray.map(function(item) {
            return {
                IPAddress: item.Base.IPAddress,
                RackNumber: item.RackNumber,
                SlotNumber: item.SlotNumber,
            };
        });
    }
    else {
        connects = [{
            IPAddress:cfg.ConnectionSettings.S7Connection.Base.IPAddress,
            RackNumber:cfg.ConnectionSettings.S7Connection.RackNumber,
            SlotNumber:cfg.ConnectionSettings.S7Connection.SlotNumber,
        }];
    }

    return connects;
}

const toViewTimeout = (cfg) => {
    // Siemens Simatic S7-400H
    if(cfg.ConnectionSettings.S7ConnectionsArray.length > 0 ){
        return cfg.ConnectionSettings.S7ConnectionsArray[0].Base.Base.Timeout;
    }
    return cfg.ConnectionSettings.S7Connection.Base.Base.Timeout;
}

const PlcRegistryCfgToViewPlc = (regCfg) => {
    const custom = regCfg.KLPLC_Custom;
    const id = Buffer.from(regCfg.KLPLC_BinId.value, 'base64');
    return {
        DisplayName: regCfg.KLPLC_DisplayName,
        ControllerType: custom.ControllerType,
        Description: regCfg.KLPLC_Descr !== '' ? regCfg.KLPLC_Descr : regCfg.KLPLC_DisplayName,
        InternalNetworkId: [].slice.call(id),
        IPAddress:custom.IPAddress,
        PortNumber:custom.PortNumber,
        RackNumber:custom.RackNumber,
        SlotNumber:custom.SlotNumber,
        ReadDBBlocks:custom.ReadDBBlocks,
        PasswordAvailable:custom.PasswordAvailable,
        NewPassword:'',
        OldPassword: (custom.Password!==undefined) ? custom.Password.value : '',
        Timeout: custom.Timeout > 0 ? custom.Timeout : 1,
        NetId:[],
        ConnectionsArray: custom.ReservedConnectios ? custom.ReservedConnectios : [],
        FromPlcRegitry: true,
        IsSiemensPlc: IsSiemensPlc(custom.ControllerType),
        hasRackSlotInput: hasRackSlotInput(custom.ControllerType)
    };
}

const PlcDevApiCfgToViewPlc = (regCfg, internalId) => {
    const custom = regCfg.KLPLC_Custom.value;
    return {
        DisplayName: regCfg.KLPLC_DisplayName,
        ControllerType: custom.ControllerType,
        Description: regCfg.KLPLC_Descr !== '' ? regCfg.KLPLC_Descr : regCfg.KLPLC_DisplayName,
        InternalNetworkId: internalId,
        IPAddress:custom.IPAddress,
        PortNumber:custom.PortNumber,
        RackNumber:custom.RackNumber,
        SlotNumber:custom.SlotNumber,
        ReadDBBlocks:custom.ReadDBBlocks,
        PasswordAvailable:custom.PasswordAvailable,
        NewPassword:'',
        OldPassword: (custom.Password!==undefined) ? custom.Password.value : '',
        Timeout: custom.Timeout > 0 ? custom.Timeout : 1,
        NetId:[],
        ConnectionsArray: custom.ReservedConnectios ? custom.ReservedConnectios : [],
        FromPlcRegitry: true,
        IsSiemensPlc: IsSiemensPlc(custom.ControllerType),
        hasRackSlotInput: hasRackSlotInput(custom.ControllerType)
    };
}

const GenerateInternalPlcId = (theDate) => {
    const curDate = theDate ? theDate : new Date();
    let out = pad(curDate.getFullYear(), 4);
    out += pad(curDate.getMonth()+1, 2);
    out += pad(curDate.getDate(), 2);
    out += pad(curDate.getHours(), 2);
    out += pad(curDate.getMinutes(), 2);
    out += pad(curDate.getSeconds(), 2);
    const ms = curDate.getMilliseconds() >= 100 ? (curDate.getMilliseconds() / 10) : curDate.getMilliseconds();
    out += pad(ms, 2);
    return out.substr(0,16);
}

const MakeHexNetId = (NetId) => {
    const HexCharsW  = Array.from('0123456789ABCDEF');
    let result = '0x';
    for (let i = 0; i < NetId.length; i++) {
        result += HexCharsW[(NetId[i] >> 4) & 0xF];
        result += HexCharsW[NetId[i] & 0xF];
    }
    return result;
}

const toHexString = (byteArray) => {
    return byteArray.reduce((output, elem) => (output + ('0' + elem.toString(16)).slice(-2)), '');
}

const hexToBytes = (hex) => {
    for (let bytes = [], c = 0; c < hex.length; c += 2)
        bytes.push(parseInt(hex.substr(c, 2), 16));
    return bytes;
}

const toRowPlcRegistryCfg = (viewCfg, cryptPswrd) => {
    let custom = {
        ControllerType: viewCfg.ControllerType,
        IPAddress: viewCfg.IPAddress,
        PasswordAvailable: viewCfg.PasswordAvailable,
        PortNumber: viewCfg.PortNumber,
        RackNumber: viewCfg.RackNumber,
        ReadDBBlocks: viewCfg.ReadDBBlocks,
        SlotNumber: viewCfg.SlotNumber,
        Timeout: viewCfg.Timeout,
        Password: { type: 'binary', value: viewCfg.PasswordAvailable ? cryptPswrd : '' }
    };
    if(viewCfg.ControllerType===8 && viewCfg.ConnectionsArray.length > 0 ){
        custom.IPAddress = viewCfg.ConnectionsArray[0].IPAddress;
        custom.RackNumber = viewCfg.ConnectionsArray[0].RackNumber;
        custom.SlotNumber = viewCfg.ConnectionsArray[0].SlotNumber;
        custom.ReservedConnectios = viewCfg.ConnectionsArray.map((item)=>{
            return {
                IPAddress: item.IPAddress,
                RackNumber: item.RackNumber,
                SlotNumber: item.SlotNumber,
            };
        });
    }
    if(viewCfg.HostsList){
        custom.HostsList = viewCfg.HostsList;
    }

    const binId = (viewCfg.BinInternalId) ? viewCfg.BinInternalId : btoa(GenerateInternalPlcId());
    const id = GetPlcId(viewCfg);
    return {
        KLPLC_BinId: { type: 'binary', value: binId },
        KLPLC_ConnectionString: getPlcConnDescription(viewCfg) + '; id: ' + id,
        KLPLC_Custom: custom,
        KLPLC_Descr: viewCfg.Description,
        KLPLC_DisplayName: viewCfg.DisplayName ? viewCfg.DisplayName : viewCfg.Description,
        KLPLC_Model: GetLocalizedPlcType(viewCfg.ControllerType),
        KLPLC_Publisher: GetLocalizedPlcPublisher(viewCfg.ControllerType),

    };
}

const DefaultNListPlcConfig = () => {
    return {
        Enabled: true,
        Firmwares: [],
        PollingMinutes: 1,
        PollingSeconds: 0,
        ConnectionSettings: {},
        ControllerType: 0,
        Description: '',
        IPAddress:'',
        PortNumber:0,
        RackNumber:0,
        SlotNumber:0,
        ReadDBBlocks:false,
        PasswordAvailable:false,
        OldPassword:'',
        ViewConnectionsArray:[],
        InternalNetworkId: [],
        NetId: [],
        EditOnly:false
    };
};

const ViewNListPlcConfig = (nlistCfg) => {
    const cfg = nlistCfg.PLCConfig;
    return {
        Enabled: true,
        Firmwares: [], //cfgChecker.Firmwares,
        PollingMinutes: 1,
        PollingSeconds: 0,
        ConnectionSettings: cfg.ConnectionSettings,
        ControllerType: cfg.ControllerType,
        Description: nlistCfg.PLCName!=='' ? nlistCfg.PLCName : cfg.Description,
        IPAddress:cfg.ConnectionSettings.S7Connection.Base.IPAddress,
        PortNumber:cfg.ConnectionSettings.S7Connection.Base.PortNumber,
        RackNumber:cfg.ConnectionSettings.S7Connection.RackNumber,
        SlotNumber:cfg.ConnectionSettings.S7Connection.SlotNumber,
        ReadDBBlocks:cfg.ConnectionSettings.S7Connection.ReadDBBlocks,
        PasswordAvailable:cfg.ConnectionSettings.S7Connection.Base.Base.PasswordAvailable,
        OldPassword:cfg.ConnectionSettings.S7Connection.Base.Base.Password.value,
        ViewConnectionsArray: DisplayedCheckConnectionArray(cfg),
        InternalNetworkId: cfg.InternalNetworkId.InternalNetworkId,
        NetId: cfg.NetworkId.NetId,
        EditOnly:false
    };
};

const ViewFirmware = (firmware) => {
    const ts = Math.floor(Date.parse(firmware.HistDetect.value)/1000);
    return {
        Created: {type: 'long',value: ts.toString()},
        FirmwareValue: hexToBytes(firmware.FB),
        Info: firmware.FD,
        __VersionInfo: [1,0]
    }
}

const IsPlcDublicated = (plc, vPlcCfgs) => {
    const PlcId = GetPlcId(plc);
    const index = vPlcCfgs.findIndex((cfg) => {
        return PlcId === GetPlcId(cfg);
    });
    return ( index >= 0 ) ? true : false;
}

const IsRegistryPlcEqual = (plc1, plc2) => {
    if(plc1.KLPLC_Custom.ControllerType !== plc2.KLPLC_Custom.ControllerType )
        return false;

    if(plc1.KLPLC_Custom.IPAddress !== plc2.KLPLC_Custom.IPAddress ||
        plc1.KLPLC_Custom.PortNumber !== plc2.KLPLC_Custom.PortNumber ||
         plc1.KLPLC_Custom.RackNumber !== plc2.KLPLC_Custom.RackNumber ||
          plc1.KLPLC_Custom.SlotNumber !== plc2.KLPLC_Custom.SlotNumber ||
           plc1.KLPLC_Custom.ReadDBBlocks !== plc2.KLPLC_Custom.ReadDBBlocks )
        return false;

    if(plc1.KLPLC_Custom.ReservedConnectios || plc2.KLPLC_Custom.ReservedConnectios){
        const a1 = JSON.stringify(plc1.KLPLC_Custom.ReservedConnectios);
        const a2 = JSON.stringify(plc2.KLPLC_Custom.ReservedConnectios);
        return a1 === a2;
    }

    return true;
}

const IsConnectionDublicated = (model, parentFormId) => {
    const parentForm = runtime.ui.form(parentFormId);
    if(!parentForm)
        return false;
    const modelParent = parentForm.getModel();
    if(!modelParent || !modelParent.ConnectionsArray)
        return false;
    const found = modelParent.ConnectionsArray.find((conn) => {
        return conn.IPAddress===model.IPAddress && conn.RackNumber===model.RackNumber && conn.SlotNumber===model.SlotNumber;
    });
    return found ? true : false;
}

const toRowXmlPlcRegistryCfg = (xmlTag) => {
    const xmlCfg = xmlTag['$'];
    let custom = {
        ControllerType: parseInt(xmlCfg.Type),
        IPAddress: xmlCfg.Ip,
        PortNumber: parseInt(xmlCfg.Port),
        RackNumber: parseInt(xmlCfg.Rack),
        ReadDBBlocks: xmlCfg.ReadBlocks === '1' ? true : false,
        SlotNumber: parseInt(xmlCfg.Slot),
        Timeout: parseInt(xmlCfg.Timeout),
        PasswordAvailable: false,
        Password: { type: 'binary', value: '' },
    };
    if(xmlTag['Conect'] && xmlTag['Conect'].length > 0 ){
        const connects = xmlTag['Conect'];
        custom.IPAddress = Array.isArray(connects[0].IPAddress) ? connects[0].IPAddress[0] : connects[0].IPAddress;
        custom.RackNumber = parseInt(connects[0].RackNumber[0]);
        custom.SlotNumber = parseInt(connects[0].SlotNumber[0]);

        custom.ReservedConnectios = connects.map((item)=>{
            return {
                IPAddress: item.IPAddress[0],
                RackNumber: parseInt(item.RackNumber[0]),
                SlotNumber: parseInt(item.SlotNumber[0]),
            };
        });
    }
    const id = GetPlcId(custom);
    return {
        KLPLC_BinId: { type: 'binary', value: '' },
        KLPLC_ConnectionString: getPlcConnDescription(custom) + '; id: ' + id,
        KLPLC_Custom: custom,
        KLPLC_Descr: xmlCfg.Description,
        KLPLC_DisplayName: xmlCfg.Name ? xmlCfg.Name : xmlCfg.Description,
        KLPLC_Model: xmlCfg.Model,
        KLPLC_Publisher: xmlCfg.Publisher,
    };
}

const toExportPlcRegistryCfg = (cfg, id) => {
    let exportCfg = {};
    exportCfg['$'] = {
        id: pad(id, 5),
        Name: cfg.KLPLC_DisplayName,
        Description: cfg.KLPLC_Descr,
        Model: cfg.KLPLC_Model,
        Publisher: cfg.KLPLC_Publisher,
        Type: cfg.KLPLC_Custom.ControllerType.toString(),
        Ip: cfg.KLPLC_Custom.IPAddress,
        Port: cfg.KLPLC_Custom.PortNumber.toString(),
        Rack: cfg.KLPLC_Custom.RackNumber.toString(),
        Slot: cfg.KLPLC_Custom.SlotNumber.toString(),
        Timeout: cfg.KLPLC_Custom.Timeout.toString(),
        ReadBlocks: cfg.KLPLC_Custom.ReadDBBlocks ? '1' : '0'
    }
    if(cfg.KLPLC_Custom.ReservedConnectios ){
        exportCfg.Conect = cfg.KLPLC_Custom.ReservedConnectios.map((conn)=>{
            return  { IPAddress: conn.IPAddress, RackNumber: conn.RackNumber.toString(), SlotNumber: conn.SlotNumber.toString()}
        });
    }
    return exportCfg;
}

//
// import/export PlcInvestigate data
//
const xmlToViewPlcInvestigateCfg = (xmlCfg) => {
    const controllerType = parseInt(xmlCfg.ControllerType[0]);
    const connSetts = xmlCfg.ConnectionSettings[0].S7Connection[0];
    const base = xmlCfg.ConnectionSettings[0].S7Connection[0].Base[0];
    const internalId = xmlCfg.InternalNetworkId[0].InternalNetworkId[0].element ?
        xmlCfg.InternalNetworkId[0].InternalNetworkId[0].element.map((e)=>{ return parseInt(e);}) : [];

    // Siemens Simatic S7-400H
    let connects = [];
    if(controllerType === 8 ) {
        if(xmlCfg.ConnectionSettings[0].S7ConnectionsArray[0].element !== undefined ){
            connects = xmlCfg.ConnectionSettings[0].S7ConnectionsArray[0].element.map(function(item) {
                return {
                    IPAddress: item.Base[0].IPAddress[0].replace(/[\r\n\t]/g, ""),
                    RackNumber: parseInt(item.RackNumber[0]),
                    SlotNumber: parseInt(item.SlotNumber[0]),
                };
            });
        }
        else {
            connects = [{
                IPAddress: base.IPAddress[0].replace(/[\r\n\t]/g, ""),
                RackNumber: parseInt(connSetts.RackNumber[0]),
                SlotNumber: parseInt(connSetts.SlotNumber[0]),
            }];
        }
    }

    return {
        ControllerType: controllerType,
        Description: xmlCfg.Description[0].replace(/[\r\n\t]/g, ""),
        InternalNetworkId: internalId,
        IPAddress: base.IPAddress[0].replace(/[\r\n\t]/g, ""),
        PortNumber: parseInt(base.PortNumber[0]),
        RackNumber: parseInt(connSetts.RackNumber[0]),
        SlotNumber: parseInt(connSetts.SlotNumber[0]),
        ReadDBBlocks: connSetts.ReadDBBlocks[0] === 'yes',
        PasswordAvailable: false,
        NewPassword:'',
        OldPassword: '',
        Timeout: parseInt(base.Base[0].Timeout[0]),
        IsVirtualController: base.Base[0].IsVirtualController ? (base.Base[0].IsVirtualController[0]==='yes') : false,
        NetId: [],
        ConnectionsArray: connects,
        FromPlcRegitry: false,
        IsSiemensPlc: IsSiemensPlc(controllerType),
        hasRackSlotInput: hasRackSlotInput(controllerType)
    };
}

const xmlFromViewPlcInvestigateCfg = (viewCfg) => {
    const internalId = viewCfg.InternalNetworkId.length > 0 ?  { element: viewCfg.InternalNetworkId.map(i => { return i.toString()}) } : '';
    let xmlConns = [];
    if(viewCfg.ControllerType === 8) {
        xmlConns = viewCfg.ConnectionsArray.map(item => {
            return {
                __VersionInfo:[{element:["1","0"]}],
                Base:[{ __VersionInfo:[{element:["1","0"]}],
                    Base:[{
                        __VersionInfo:[{element:["1","1"]}],
                        Timeout:[viewCfg.Timeout.toString()],
                        PasswordAvailable:["no"],
                        Password:[""],
                        IsVirtualController:[viewCfg.IsVirtualController ? "yes" : "no"]
                    }],
                    IPAddress:[item.IPAddress.toString()],
                    PortNumber:[viewCfg.PortNumber.toString()]
                }],
                RackNumber:[item.RackNumber.toString()],
                SlotNumber:[item.SlotNumber.toString()],
                ReadDBBlocks:[viewCfg.ReadDBBlocks ? "yes" : "no"]
            }
        });
    }

    return {
        __VersionInfo: [{element:["1","1"]}],
        NetworkId: [{__VersionInfo: [{element:["1","0"]}], NetId: [""] }],
        ControllerType: [viewCfg.ControllerType.toString()],
        Description: [viewCfg.Description],
        ConnectionSettings: [{ __VersionInfo: [{element:["1","1"]}],
            S7Connection: [{ __VersionInfo: [{element:["1","0"]}],
                Base: [{ __VersionInfo: [{element:["1","0"]}],
                    Base: [{ __VersionInfo: [{element:["1","1"]}],
                        Timeout: [viewCfg.Timeout.toString()],
                        PasswordAvailable: ["no"],
                        Password: [""],
                        IsVirtualController: [viewCfg.IsVirtualController ? "yes" : "no"]
                    }],
                    IPAddress: [viewCfg.IPAddress],
                    PortNumber: [viewCfg.PortNumber.toString()]
                }],
                RackNumber: [viewCfg.RackNumber.toString()],
                SlotNumber: [viewCfg.SlotNumber.toString()],
                ReadDBBlocks: [viewCfg.ReadDBBlocks ? "yes" : "no"]
            }],
            S7ConnectionsArray: [ xmlConns.length > 0 ? {element: xmlConns} : '' ]
        }],
        InternalNetworkId: [{ __VersionInfo: [{element:["1","0"]}], InternalNetworkId:[internalId] }]
    };
}

module.exports = {
    GetLocalizedPlcType,
    GetLocalizedPlcPublisher,
    IsSiemensPlc,
    GetPlcId,
    hasRackSlotInput,
    getPlcConnDescription,
    toViewConnectionArray,
    DisplayedCheckConnectionArray,
    toViewTimeout,
    PlcRegistryCfgToViewPlc,
    PlcDevApiCfgToViewPlc,
    toRowPlcRegistryCfg,
    GenerateInternalPlcId,
    MakeHexNetId,
    toHexString,
    DefaultNListPlcConfig,
    ViewNListPlcConfig,
    ViewFirmware,
    IsPlcDublicated,
    IsRegistryPlcEqual,
    IsConnectionDublicated,
    toRowXmlPlcRegistryCfg,
    toExportPlcRegistryCfg,
    // import/export PlcInvestigate data
    xmlToViewPlcInvestigateCfg,
    xmlFromViewPlcInvestigateCfg
}