const buffer = require('buffer');
const binContainer = require('./binaryContainer.js');
const serTypes = require('./serialize_types.js');

// from \development\common\appinfo\1810r.h
const COMPATIBLE_APP_ID = [1810,1555,1828];
const ChunkSize = 500;

/*
 * Common methods
 */
async function IsSetHostedMode(kscApi) {
  if(kscApi.utils!==undefined && kscApi.utils.getWebConsoleInfo!==undefined) {
        const info = await kscApi.utils.getWebConsoleInfo();
        return info.isHosted!==undefined ? info.isHosted : false;
    }
    return false;
}
/*
 * Ksn / Kpsn api
 */
async function IsSetKpsnMode(kscApi) {
    const eulas = await kscApi.KsnInternal.GetNKsnEulas();
    return (eulas.result!==undefined && eulas.result.pNKsnEulas!==undefined && eulas.result.pNKsnEulas.length > 0 ) ? true : false;
}

/*
 * Licensing api
 */
const GetKeyData = async (kscApi, serial) => {
    const response = await kscApi.LicenseKeys.GetKeyData({
        pKeyInfo: {'KLLIC_SERIAL': serial, 'KLLICSRV_KEYDATA': true, 'KLLIC_APP_IDS': true}
    })
    return response.result;
}

async function EnumCompatibleLicenseKeys(kscApi) {
    const licFields = ['KLLIC_APP_ID', 'KLLIC_LIMIT_DATE','KLLIC_KEY_TYPE','KLLIC_SERIAL','KLLICSRV_KEY_INSTALLED'];
    const response = await kscApi.LicenseKeys.EnumKeys({ pFields: licFields });
    const enumlics = response.result.filter(
        lic => lic.KLLICSRV_KEY_INSTALLED===true && new Date(lic.KLLIC_LIMIT_DATE.value) > new Date(),
    );

    const productKeys = await Promise.all(
        enumlics.map(( license ) => GetKeyData(kscApi, license.KLLIC_SERIAL))
    );
    const compatibleLics = productKeys.filter(lic => {
        for(let appId of Object.keys(lic.KLLIC_APP_IDS)){
            if(COMPATIBLE_APP_ID.includes(parseInt(appId)))
                return true;
        }
        return false;
    });

    return compatibleLics;
}

/*
 * SrvView api
 */
const EnumSrvViewRecords = async (kscApi, wsViewName, wsFilter, vFields) => {
    let recs = [];
    try {
        const iterator = await kscApi.SrvView.ResetIterator({wstrViewName: wsViewName, wstrFilter: wsFilter, vecFieldsToReturn: vFields});
        if (!iterator.result)
            return recs;

        const nRecords = await kscApi.SrvView.GetRecordCount({wstrIteratorId:iterator.result});

        let nFirst = 0;
        while(nFirst < nRecords.result) {
            const nLast = (nFirst + ChunkSize) < nRecords.result  ? nFirst + ChunkSize : nRecords.result;
            const recs_range = await kscApi.SrvView.GetRecordRange({ wstrIteratorId:iterator.result, nStart: nFirst, nEnd: nLast});
            if(recs_range.result.KLCSP_ITERATOR_ARRAY) {
                recs = recs.concat(recs_range.result.KLCSP_ITERATOR_ARRAY);
                nFirst += recs_range.result.KLCSP_ITERATOR_ARRAY.length;
            }
            else
                break;
        }

        await kscApi.SrvView.ReleaseIterator({wstrIteratorId:iterator.result});

    } catch (e) {
        runtime.logger.log('- kess_windows - : EnumSrvViewRecords failture. Error = ' + e);
    }
    return recs;
}

/*
 * Plc management api
 */

const IsCompatibleWithKics4Nodes = (plcCfg) => {
    if(plcCfg===undefined || plcCfg.KLPLC_Custom===undefined || plcCfg.KLPLC_Custom.ControllerType===undefined)
        return false;
    return (plcCfg.KLPLC_Custom.ControllerType > 0 && plcCfg.KLPLC_Custom.ControllerType < 9) ? true : false;
}

async function EnumRegistryPlcConfigs(kscApi, vUsedPlcIds) {
    const vFields = ['KLPLC_BinId', 'KLPLC_DisplayName', 'KLPLC_Publisher', 'KLPLC_Model', 'KLPLC_Descr',
        'KLPLC_ConnectionString', 'KLPLC_Custom'];
    const cfgs = await EnumSrvViewRecords(kscApi, 'PLCDevicesSrvViewName','', vFields);
    const freeCfgs = cfgs.filter((x)=>{
        if(!IsCompatibleWithKics4Nodes(x))
            return false;
        const nId = Buffer.from(x.KLPLC_BinId.value, 'base64').toString();
        return vUsedPlcIds.some(function(x){ return (x === nId) }) ? false : true;
    });
    return freeCfgs;
}

async function EnumNetworkListPlcConfigs(kscApi, hostId, vUsedNetIds) {
    const vFields = ['HostId', 'HostDN', 'PLCId', 'PLCName', 'PLCAdditionalInfo'];
    const sFilter = '(HostId="' + hostId + '")';
    const cfgs = await EnumSrvViewRecords(kscApi, 'HstCipDev',sFilter, vFields);

    const freeCfgs = cfgs.filter((x)=>{
        const nId = Buffer.from(x.PLCId.value, 'base64').toString();
        return vUsedNetIds.some((x)=>{ return (x === nId) }) ? false : true;
    });

    const viewCfgs = freeCfgs.map((x)=>{
        const data = Buffer.from(x.PLCAdditionalInfo.value, 'base64');
        const PlcAdditionalInfo = binContainer.deserialize([].slice.call(data), serTypes.PLCInvestigateResultDT);
        return { PLCId: x.PLCId, PLCName : x.PLCName, HostId: x.HostId, ...PlcAdditionalInfo };
    });

    return viewCfgs;
}

async function EnumNetworkListFirmwares(kscApi, NetId) {
    const vFields = ['HistItemId', 'HistDetect', 'HistBlob'];
    const sFilter = '(HistPLCId=' + NetId + ')';
    const firmwares = await EnumSrvViewRecords(kscApi, 'CipDevHistory',sFilter, vFields);

    const viewFirmwares = firmwares.map((x)=>{
        const data = binContainer.blobToJson(x.HistBlob.value);
        data.FB = data.FB.substring(4);
        return { HistDetect: x.HistDetect, HistItemId: x.HistItemId, ...data };
    });
    return viewFirmwares;
}

const GetPlcFromRegistry = async (kscApi, internalId) => {
    try{
        const res = await kscApi.PLCDevApi.GetPLC ( { pPLCId: internalId });
        if(res.result!==undefined && res.result.PxgError!==undefined ){
            return {};
        }
        return res.result;

    } catch (e) {
        runtime.logger.log('- kess_windows - : GetPlcFromRegistry failture. Error = ' + e);
    }
    return {};
}

const UpdatePlcFromRegistry = async (kscApi, regCfg) => {
    const internalId = regCfg.KLPLC_BinId.value;
    try{
        const res = await kscApi.PLCDevApi.UpdatePLC( { pPLCId: internalId, pPLCInfo: regCfg });
        if(res.result!==undefined && res.result.PxgError!==undefined ){
            return [];
        }
    } catch (e) {
        runtime.logger.log('- kess_windows - : AddPlcToRegistry failture. Error = ' + e);
        return [];
    }
    const id = Buffer.from(internalId, 'base64');
    return [].slice.call(id);
}

const DeletePlcFromRegistry = async (kscApi, internalId) => {
    try {
        const res = await kscApi.PLCDevApi.DeletePLC( { pPLCId: internalId });
        return (res.result!==undefined && res.result.PxgError!==undefined ) ? false : true;
    } catch (e) {
        runtime.logger.log('- kess_windows - : DeletePlcFromRegistry failture. Error = ' + e);
    }
    const id = Buffer.from(internalId, 'base64');
    return false;
}

/*
 * Wifi management api
 */
async function EnumNetworkListWiFi(kscApi) {
    const vFields = ['dev_id', 'dev_type', 'dev_name'];
    const sFilter = '(&(dev_type=13))';
    return await EnumSrvViewRecords(kscApi, 'DCDevicesList', sFilter, vFields);
}

/*
 * Password management api
 */
const CryptPassword = async (kscApi, pswrd) => {
    if(pswrd === '')
        return '';
    let bytes = Buffer.from(pswrd + String.fromCharCode(0), 'utf-8');
    let bufStr16 = buffer.transcode(bytes, 'utf8', 'utf16le');
    const pswrdCmdExec = bufStr16.toString('base64');
    const response = await kscApi.DataProtectionApi.ProtectDataGlobally({ pData:pswrdCmdExec })
    return response.result;
}

module.exports = {
    IsSetKpsnMode,
    IsSetHostedMode,

    EnumCompatibleLicenseKeys,
//    EnumSrvViewRecords,
    // Plc methods
    EnumRegistryPlcConfigs,
    EnumNetworkListPlcConfigs,
    EnumNetworkListFirmwares,
    GetPlcFromRegistry,
    UpdatePlcFromRegistry,
    DeletePlcFromRegistry,
    // WiFi
    EnumNetworkListWiFi,

    CryptPassword
}