const crypto = require('crypto');


const CACHE_SIZE = 1024;
const DEFAULT_SID_FILTER = '(ul_imgSid)';
const DEFAULT_FIELDS = ['ul_wstrDisplayName', 'ul_imgSid', 'ul_binSidHash', 'ul_bUser', 'ul_wstrDomainSAM'];
const DEFAULT_ORDER = [
  {
    type: 'params',
    value: { Name: 'ul_wstrDomainSAM', Asc: true, },
  },
  {
    type: 'params',
    value: { Name: 'ul_wstrDisplayName', Asc: true, },
  },
];

/**
 * In-place cache service implementation
 * @param {number} maxSize
 */
const createLRUCache = (maxSize) => {
  const cache = new Map();
  const removeLast = () => {
    cache.delete(cache.keys().next().value);
  };
  return {
    add: (key, data) => {
      for (let i = cache.size - maxSize; i > 0; i--) {
        removeLast();
      }
      cache.set(key, data);
    },
    get: (key) => {
      const item = cache.get(key);
      if (cache.has(key)) {
        cache.delete(key);
        cache.set(key, item);
      }
      return item;
    },
  };
};

const usersCache = createLRUCache(CACHE_SIZE);

const calculateSidHash = (sid) => {
  const md5 = crypto.createHash('md5').update(Buffer.from(sid, 'base64')).digest();
  return md5.toString('hex');
}

const FilterBySid = (sids) => {
  const filters = sids.map(sid => {
    const hex = calculateSidHash(sid);
    return `(ul_binSidHash = 0x${hex})`;
  } ).join('');
  return filters;
}

const EnumKscUsers = async (kscApi, startPos, endPos, filter) => {
  try {
    const iterator = await kscApi.SrvView.ResetIterator({
      wstrViewName: 'GlobalUsersListSrvViewName',
      wstrFilter: filter,
      vecFieldsToReturn: DEFAULT_FIELDS
    });
    if (!iterator.result)
      return { rows: [], count: 0 };

    const nRecords = await kscApi.SrvView.GetRecordCount({wstrIteratorId:iterator.result});
    if (startPos >= nRecords.result){
      await kscApi.SrvView.ReleaseIterator({wstrIteratorId:iterator.result});
      return { rows: [], count: nRecords.result };
    }

    const recsRange = await kscApi.SrvView.GetRecordRange({
      wstrIteratorId: iterator.result,
      nStart: startPos,
      nEnd: (endPos < nRecords.result  ? endPos : nRecords.result)
    });
    await kscApi.SrvView.ReleaseIterator({wstrIteratorId:iterator.result});

    if(recsRange.result.KLCSP_ITERATOR_ARRAY) {
      return { rows: recsRange.result.KLCSP_ITERATOR_ARRAY, count:nRecords.result};
    }
  } catch (e) {
    runtime.logger.log('- kess_windows - : EnumSrvViewRecords failture. Error = ' + e);
  }
  return { rows: [], count: 0 };
}

const SearchKscUsers = async (kscApi, startPos, endPos, search, ignoreSids = []) => {
  let filter = DEFAULT_SID_FILTER;
  const ignoreFilters = ignoreSids.length > 0 ? FilterBySid(ignoreSids) : '';
  if (search) {
    const escapedSearch = String(search.replace('"', '\\22'));
    const compositeFilter = /^(.+?)\\(.+)$/.test(escapedSearch)
        ? escapedSearch.replace(/^(.+?)\\(.+)$/, '(&( ul_wstrDisplayName = "$2*" )( ul_wstrDomainSAM = "$1" ))')
        : '';

    filter = `(&${DEFAULT_SID_FILTER}(|( ul_wstrDisplayName = "*${escapedSearch}*" )( ul_wstrDomainSAM = "*${escapedSearch}*" )${compositeFilter})${ignoreFilters})`;
  } else if (ignoreFilters) {
    filter = `(&${DEFAULT_SID_FILTER}${ignoreFilters})`;
  }

  const records = await EnumKscUsers(kscApi, startPos, endPos, filter);
  return records;
}

async function GetKscUsersBySid(kscApi, sids) {
  const Filters = sids.length > 0 ? FilterBySid(sids) : DEFAULT_SID_FILTER;
  const records = await EnumKscUsers(kscApi, 0, 10000000, `(&(ul_imgSid)(|${Filters}))`);
  return records.rows;
}

module.exports = {
  SearchKscUsers,
  GetKscUsersBySid,
}
