const lo = require('lodash');
const hlpDef = require('../default/policy.js');

const OriginTypeEnum = { '':0, 'Default':1, 'User': 2, 'Auto':3 };
const TargetTypeEnum = {'Binary':1, 'ScriptOrMsi': 2};
const ActionTypeEnum = { 'Allow':1, 'Deny':2, 'Audit':3};
const ConditionTypeEnum = { 'Certificate': 1, 'Sha256': 2, 'Path': 3 };
const BinaryEngineType	= 0x80000000;

const EMPTY_THUMBPRINT = '0000000000000000000000000000000000000000';
const EMPTY_SHA256 = '0000000000000000000000000000000000000000000000000000000000000000';
const EMPTY_CERT_CONDITION = {UseSubject: 'false',  Subject: '',  UseThumbprint: 'false', Thumbprint: ''};
const HashPrefixRegex = /(0x{1,64})/;

const charsToRemove = /[\u007f-\u00bf]/g;

const NormalizePepHash = (hash) => {
    if(hash && hash !== '') {
        return HashPrefixRegex.test(hash) ? hash : ('0x' + hash);
    }
    return hash;
};

const getUiCertCondition = (item) => {
    const data = lo.get(item, 'CertificateRuleCondition.0.$', EMPTY_CERT_CONDITION);
    return {
        Use: (data.UseSubject === 'true' || data.UseThumbprint === 'true'),
        UseSubject: data.UseSubject === 'true',
        Subject: data.Subject,
        UseThumbprint: data.UseThumbprint === 'true',
        Thumbprint: data.Thumbprint
    };
};

const getUiHashCondition = (item) => {
    const hash = lo.get(item, 'HashRuleCondition.0.$.Hash', '');
    return { Use: hash !== '', Hash: hash };
};

const getUiPathCondition = (item) => {
    const path = lo.get(item, 'PathRuleCondition.0.$.Path', '');
    const cmd = lo.get(item, 'PathRuleCondition.0.$.CmdLine', '');
    return { Use: path !== '' || cmd !== '', Path: path, CmdLine: cmd };
};

const getUiCertExclusion = (item) => {
    const source = item.$;
    return {
        RuleCondition: ConditionTypeEnum.Certificate,
        CertificateRuleCondition: {
            UseSubject: source.UseSubject === 'true',
            Subject: source.Subject,
            UseThumbprint: source.UseThumbprint === 'true',
            Thumbprint: source.Thumbprint
        },
        HashRuleCondition: {Hash: ""},
        PathRuleCondition: {Path: ""}
    };
};

const getUiHashExclusion = (item) => {
    return {
        RuleCondition: ConditionTypeEnum.Sha256,
        CertificateRuleCondition: { UseSubject: false, Subject: "", UseThumbprint: false, Thumbprint:"" },
        HashRuleCondition: {Hash: lo.get(item.$, "Hash", "")},
        PathRuleCondition: {Path: ""}
    };
};

const getUiPathExclusion = (item) => {
    return {
        RuleCondition: ConditionTypeEnum.Path,
        CertificateRuleCondition: { UseSubject:false, Subject: "", UseThumbprint:false, Thumbprint:"" },
        HashRuleCondition: {Hash: ""},
        PathRuleCondition: {Path: lo.get(item.$, "Path", "")}
    };
};

function AddUiExclusionByType(ExclCollection, exclType, uiExcl) {
    if(ExclCollection[exclType]===undefined )
        return;
    for(let item of ExclCollection[exclType]) {
        if(exclType === 'CertificateRuleCondition') {
            uiExcl.push(getUiCertExclusion(item));
        } else if(exclType === 'HashRuleCondition') {
            uiExcl.push(getUiHashExclusion(item));
        } else if(exclType === 'PathRuleCondition') {
            uiExcl.push(getUiPathExclusion(item));
        }
    }
};

const toUiRuleExclusions = (jsonExclusions) =>{
    let exclusions = [];
    if(jsonExclusions===undefined || jsonExclusions[0]===undefined)
        return exclusions;
    AddUiExclusionByType(jsonExclusions[0], 'CertificateRuleCondition', exclusions);
    AddUiExclusionByType(jsonExclusions[0], 'HashRuleCondition', exclusions);
    AddUiExclusionByType(jsonExclusions[0], 'PathRuleCondition', exclusions);
    return exclusions;
};

const toUiRules = (item) => {
    const name = lo.get(item.$, "Name", "");
    let cert = getUiCertCondition(item);
    const hash = getUiHashCondition(item);
    const path = getUiPathCondition(item);

    let condition = ConditionTypeEnum.Path;
    if( hash.Use ){
        condition = ConditionTypeEnum.Sha256;
    }
    if( cert.Use ){
        condition = ConditionTypeEnum.Certificate;
    }
    // set as cert rule when all conditions unused
    if( !(cert.Use || hash.Use || path.Use) ){
        condition = ConditionTypeEnum.Certificate;
        cert.Use = true;
    }

    return {
        Name: name.replace(charsToRemove, ''),
        Origin: OriginTypeEnum[lo.get(item.$, "Origin", "Default")],
        Target: TargetTypeEnum[lo.get(item.$, "Target", "ScriptOrMsi")],
        Action: ActionTypeEnum[lo.get(item.$, "Action", "Allow")],
        UserOrGroupSid: lo.get(item.$, "UserOrGroupSid", 'S-1-1-0'),
        RuleCondition: condition,
        CertificateRuleCondition: cert,
        HashRuleCondition: hash,
        PathRuleCondition: path,
        Exclusions: toUiRuleExclusions(item.Exclusions)
    };
};

const getRuleAttributes = (uiRule) => {
    const OriginNamesByType = { 0:'', 1:'Default', 2:'User', 3:'Auto' };
    const TargetNamesByType = { 1:'Binary', 2:'ScriptOrMsi'};
    const ActionNamesByType = { 1:'Allow', 2:'Deny', 3:'Audit'};
    return {
        "Name": uiRule.Name,
        "Description": "",
        "Origin": OriginNamesByType[uiRule.Origin],
        "Target": TargetNamesByType[uiRule.Target],
        "UserOrGroupSid": uiRule.UserOrGroupSid,
        "Action": ActionNamesByType[uiRule.Action]
    };
};

const getCertRuleCondition = (obj) => {
    return {
        "$":{
            "UseSubject": obj.UseSubject ? "true" : "false",
            "Subject": obj.Subject,
            "UseThumbprint": obj.UseThumbprint ? "true" : "false",
            "Thumbprint": obj.Thumbprint
        }
    };
};

const getRuleExclusions = (uiExcl) => {
    let CertExcl = [];
    let HashExcl = [];
    let PathExcl = [];

    for(let item of uiExcl) {
        if(item.RuleCondition===ConditionTypeEnum.Certificate){
            const certExcl = getCertRuleCondition(item.CertificateRuleCondition);
            CertExcl.push(certExcl);
        }
        else if(item.RuleCondition===ConditionTypeEnum.Sha256){
            const hash = lo.get(item, 'HashRuleCondition.Hash', '');
            HashExcl.push({"$":{"Type":"SHA256","Hash": hash}});
        }
        else if(item.RuleCondition===ConditionTypeEnum.Path){
            const path = lo.get(item, 'PathRuleCondition.Path', '');
            PathExcl.push({"$":{"Path": path}});
        }
    }

    let excl = {};
    if(CertExcl.length){
        excl.CertificateRuleCondition = CertExcl;
    }
    if(HashExcl.length){
        excl.HashRuleCondition = HashExcl;
    }
    if(PathExcl.length){
        excl.PathRuleCondition = PathExcl;
    }

    return [excl];
};

const toJsonRule = (uiRule) => {
    let out = {
        "$": getRuleAttributes(uiRule),
    };
    if(uiRule.CertificateRuleCondition.Use) {
        out.CertificateRuleCondition = [getCertRuleCondition(uiRule.CertificateRuleCondition)];
    }
    if(uiRule.PathRuleCondition.Use && (uiRule.PathRuleCondition.Path !== '' || uiRule.PathRuleCondition.CmdLine !== '')) {
        out.PathRuleCondition = [{ "$": { "Path": uiRule.PathRuleCondition.Path, "CmdLine": uiRule.PathRuleCondition.CmdLine} }];
    }
    if(uiRule.HashRuleCondition.Use && uiRule.HashRuleCondition.Hash !== '') {
        const hash = NormalizePepHash(uiRule.HashRuleCondition.Hash);
        out.HashRuleCondition = [{ "$": { "Type": "SHA256", "Hash": hash } }];
    }

    if(uiRule.Exclusions.length)
        out.Exclusions = getRuleExclusions(uiRule.Exclusions);
    return out;
};

const getUiRuleCondition = (rule) => {
    if(rule.CertificateRuleCondition.Use)
        return ConditionTypeEnum.Certificate;
    return rule.HashRuleCondition.Use ? ConditionTypeEnum.Sha256 : ConditionTypeEnum.Path;
};

/*
* exports
*/
const toPepUiRules = (jsonRules) => {
    const collection = lo.get(jsonRules, 'ApplicationControl.RuleCollection.0',
        { CertificateRule:[], HashRule:[], PathRule:[] });
    let allRules = collection.CertificateRule ? collection.CertificateRule : [];
    allRules = allRules.concat(collection.HashRule ? collection.HashRule: []);
    allRules = allRules.concat(collection.PathRule? collection.PathRule : []);
    return allRules.map(item => {
        return toUiRules(item);
    });
};

const JsonToPepSettingsRule = (json, bHashPriority = false) => {
    try {
        const expItem = JSON.parse(json);
        if(expItem===undefined || expItem.ActionMask===undefined)
            return {};
        let rule = hlpDef.defaultApplicationControlSettingsRule();
        rule.Origin = OriginTypeEnum['Auto'];
        rule.Action = ActionTypeEnum['Allow'];
        rule.Target = (expItem.ActionMask & BinaryEngineType) ?  1 : 2;
        rule.UserOrGroupSid = 'S-1-1-0';

        const sha256 = NormalizePepHash(lo.get(expItem, "Sha256", EMPTY_SHA256));
        const path = lo.get(expItem, "ObjectName", "").replace(charsToRemove, '');

        if(expItem.Valid==='Yes' && !bHashPriority){
            const thumbLabel = expItem.ProductName !== '' ? ' Thumbprint: ' : 'Thumbprint: ';
            rule.Name = expItem.ProductName.replace(charsToRemove, '') + thumbLabel + expItem.Thumbprint;
            rule.RuleCondition = ConditionTypeEnum.Certificate;
            rule.CertificateRuleCondition = {
                Use: true,
                UseSubject: true,
                Subject: expItem.Issuer,
                UseThumbprint: true,
                Thumbprint: lo.get(expItem, "Thumbprint", EMPTY_THUMBPRINT)
            };
            rule.HashRuleCondition = {Use: (sha256 !== ''), Hash: sha256};
            rule.PathRuleCondition = {Use: (path !== ''), Path: path, CmdLine: ''};
        }
        // Create by SHA256 priority
        else{
            if( sha256 === '' || sha256 === EMPTY_SHA256 )
                return {};
            rule.Name = path.replace(/\0.*$/g,'');
            rule.RuleCondition = ConditionTypeEnum.Sha256;
            rule.HashRuleCondition = {Use: true, Hash: sha256};
            rule.CertificateRuleCondition = {...EMPTY_CERT_CONDITION, Use: false};
            rule.PathRuleCondition = {Use: false, Path: path, CmdLine: ''};
        }
        return rule;
    }
    catch (error) {
        console.error(error);
    }
    return {};
};

const toJsonRules = (uiRules, jsonTemplate) => {
    let jsonRules = jsonTemplate;

    let CertRule = [];
    let HashRule = [];
    let PathRule = [];

    for(let rule of uiRules){
        const ruleCondition = getUiRuleCondition(rule);
        if(ruleCondition === ConditionTypeEnum.Certificate){
            CertRule.push(toJsonRule(rule));
        }
        else if(ruleCondition === ConditionTypeEnum.Sha256){
            HashRule.push(toJsonRule(rule));
        }
        else if(ruleCondition === ConditionTypeEnum.Path){
            PathRule.push(toJsonRule(rule));
        }
    }
    let collect = {};
    if(CertRule.length){
        collect.CertificateRule = CertRule;
    }
    if(PathRule.length){
        collect.PathRule = PathRule;
    }
    if(HashRule.length){
        collect.HashRule = HashRule;
    }
    jsonRules.ApplicationControl.RuleCollection = [collect];
    return jsonRules;
};

const getDummyXmlRulesTemplate = (jsonRules) => {
    const ver = lo.get(jsonRules, 'ApplicationControl.$.Version', '1');
    return { ApplicationControl: { '$': {Version: ver}, 'RuleCollection': [] } }
};

const isPepRulesEqual = (baseRule, newRule) => {
    if( JSON.stringify(baseRule.CertificateRuleCondition) === JSON.stringify(newRule.CertificateRuleCondition) &&
        JSON.stringify(baseRule.HashRuleCondition) === JSON.stringify(newRule.HashRuleCondition) &&
        JSON.stringify(baseRule.PathRuleCondition) === JSON.stringify(newRule.PathRuleCondition)  &&
        baseRule.Target === newRule.Target){
        return true;
    }
    return false;
};

module.exports = {
    ConditionTypeEnum,
    toPepUiRules,
    toJsonRules,
    JsonToPepSettingsRule,
    getDummyXmlRulesTemplate,
    isPepRulesEqual,
    NormalizePepHash,
    getUiRuleCondition,
};