const hlpPlc = require('../../../common/client/plc_tools.js');
const apiUtils = require('../../../common/open_api');
const utilsCmn = require('../../utils/common.js');

const defPolicySections = [];
const updateHostsListActionEnum = {
    Add: 1,
    Remove: 2,
};

// toRow helpers
const toRowConnectionArray = (cfg, cryptedPassword) => {
    if(cfg.ControllerType !== 8) {
        return [];
    }
    const connects = cfg.ConnectionsArray.map(function(item) {
        return {
            Base: {
                Base: {
                    Password: {type: 'binary', value: cryptedPassword},
                    PasswordAvailable: cfg.PasswordAvailable,
                    Timeout: cfg.Timeout,
                    __VersionInfo: [1,0]
                },
                IPAddress: item.IPAddress,
                PortNumber: cfg.PortNumber,
                __VersionInfo: [1,0]
            },
            RackNumber: item.RackNumber,
            ReadDBBlocks: cfg.ReadDBBlocks,
            SlotNumber: item.SlotNumber,
            __VersionInfo: [1,0]
        };
    });
    return connects;
}

async function GetPlcConfig(kscApi, cfg) {
    const isPswrdAvailable = hlpPlc.IsSiemensPlc(cfg.ControllerType) ? cfg.PasswordAvailable : false;
    let cryptedPassword = isPswrdAvailable ? cfg.OldPassword : '';
    if( !cfg.FromPlcRegitry && isPswrdAvailable && cfg.NewPassword!=='') {
        const pswrdCmdExec = utilsCmn.getPswdrData(cfg.NewPassword);
        const response = await kscApi.DataProtectionApi.ProtectDataGlobally({ pData:pswrdCmdExec })
        if(response.result!=='')
            cryptedPassword = response.result;
    }

    const actualConn = (cfg.ConnectionsArray.length > 0 ) ? {
            IPAddress:  cfg.ConnectionsArray[0].IPAddress,
            RackNumber: cfg.ConnectionsArray[0].RackNumber,
            SlotNumber: cfg.ConnectionsArray[0].SlotNumber
        } : {
            IPAddress:  cfg.IPAddress,
            RackNumber: cfg.RackNumber,
            SlotNumber: cfg.SlotNumber
        };

    const connArray = toRowConnectionArray(cfg, cryptedPassword);
    return  {
        ConnectionSettings: {
            S7Connection: {
                Base: {
                    Base: {
                        Password: { type: 'binary', value: cryptedPassword },
                        PasswordAvailable: cfg.PasswordAvailable,
                        IsVirtualController: cfg.IsVirtualController ? cfg.IsVirtualController : false,
                        Timeout: cfg.Timeout,
                        __VersionInfo: [1,1]
                    },
                    IPAddress: actualConn.IPAddress,
                    PortNumber: cfg.PortNumber,
                    __VersionInfo: [1,0]
                },
                RackNumber: actualConn.RackNumber,
                ReadDBBlocks: cfg.ReadDBBlocks,
                SlotNumber: actualConn.SlotNumber,
                __VersionInfo: [1,0]
            },
            S7ConnectionsArray: connArray,
            __VersionInfo: [1,1]
        },
        ControllerType: cfg.ControllerType,
        Description: cfg.Description,
        InternalNetworkId: {InternalNetworkId: cfg.InternalNetworkId, __VersionInfo: [1,0]},
        NetworkId: { NetId: cfg.NetId, __VersionInfo: [1,0]},
        __VersionInfo: [1,1]
    };
}

const UpdateHostsListInRegistryPlc = async (kscApi, id, hostId, ActionEnum) => {
    const internalId = Buffer.from(id).toString('base64');
    let mgrCfg = await apiUtils.GetPlcFromRegistry(kscApi, internalId);
    if(!mgrCfg || mgrCfg.KLPLC_Custom===undefined )
        return;
    if(mgrCfg.KLPLC_BinId===undefined) {
        mgrCfg.KLPLC_BinId = { type: 'binary', value: internalId };
    }

    let hList = mgrCfg.KLPLC_Custom.value.HostsList  ? mgrCfg.KLPLC_Custom.value.HostsList : [];
    if(ActionEnum === updateHostsListActionEnum.Add && !hList.includes(hostId)) {
        hList.push(hostId);
        mgrCfg.KLPLC_Custom.value.HostsList = hList;
        await apiUtils.UpdatePlcFromRegistry(kscApi, mgrCfg);
    }
    else if(ActionEnum === updateHostsListActionEnum.Remove && hList.includes(hostId)) {
        mgrCfg.KLPLC_Custom.value.HostsList = hList.filter(e => { e !== hostId });
        if(mgrCfg.KLPLC_Custom.value.HostsList.length === 0 ){
            delete mgrCfg.KLPLC_Custom.value.HostsList;
        }
        await apiUtils.UpdatePlcFromRegistry(kscApi, mgrCfg);
    }
}

const UpdatePlcManagerLinks = async (kscApi, hostId, InternalIdsOnStart, InternalIdsOnEnd) => {
    // add new hostId links to registry configurations
    for(const id of InternalIdsOnEnd) {
        if(!InternalIdsOnStart.includes(id)) {
            await UpdateHostsListInRegistryPlc(kscApi, id, hostId, updateHostsListActionEnum.Add);
        }
    }
    // delete unused hostId links from registry configurations
    for(const id2 of InternalIdsOnStart){
        if(!InternalIdsOnEnd.includes(id2)) {
            await UpdateHostsListInRegistryPlc(kscApi, id2, hostId, updateHostsListActionEnum.Remove);
        }
    }
}

const toView = (TASK_ADDITIONAL_PARAMS, activePolicyCache, formType) => {
    if(formType==='Main') {
        let vUsedInternalIds = [];
        const cfgs = TASK_ADDITIONAL_PARAMS.Settings.PLCConfigs.map(function(cfg) {
            if(cfg.InternalNetworkId.InternalNetworkId.length > 0 ){
                vUsedInternalIds.push(Buffer.from(cfg.InternalNetworkId.InternalNetworkId).toString());
            }

            return {
                ControllerType:cfg.ControllerType,
                Description:cfg.Description,
                InternalNetworkId:cfg.InternalNetworkId.InternalNetworkId,
                IPAddress:cfg.ConnectionSettings.S7Connection.Base.IPAddress,
                PortNumber:cfg.ConnectionSettings.S7Connection.Base.PortNumber,
                RackNumber:cfg.ConnectionSettings.S7Connection.RackNumber,
                SlotNumber:cfg.ConnectionSettings.S7Connection.SlotNumber,
                ReadDBBlocks:cfg.ConnectionSettings.S7Connection.ReadDBBlocks,
                PasswordAvailable:cfg.ConnectionSettings.S7Connection.Base.Base.PasswordAvailable,
                NewPassword:'',
                OldPassword:cfg.ConnectionSettings.S7Connection.Base.Base.Password.value,
                IsVirtualController: cfg.ConnectionSettings.S7Connection.Base.Base.IsVirtualController,
                Timeout: hlpPlc.toViewTimeout(cfg),
                NetId:cfg.NetworkId.NetId,
                ConnectionsArray: hlpPlc.toViewConnectionArray(cfg),
                FromPlcRegitry: cfg.InternalNetworkId.InternalNetworkId.length > 0 ? true : false,
                IsSiemensPlc: hlpPlc.IsSiemensPlc(cfg.ControllerType),
                hasRackSlotInput: hlpPlc.hasRackSlotInput(cfg.ControllerType)
            };
        });

        return { PLCConfigs: cfgs, UsedInternalIds: vUsedInternalIds };
    }
    return {};
}

async function toRawEx(kscApi, TASK_ADDITIONAL_PARAMS, changedData, formType, hostId) {
    if(formType!=='Main')
        return;

    const cfgs = await Promise.all(
        changedData.PLCConfigs.map(( cfg ) => GetPlcConfig(kscApi, cfg))
    );
    TASK_ADDITIONAL_PARAMS.Settings.PLCConfigs = cfgs;

    // async update Plc manager configures (section HostsList)
    let InternalIdsOnEnd = [];
    changedData.PLCConfigs.forEach((cfg)=>{
        if(cfg.InternalNetworkId.length > 0 ){
            InternalIdsOnEnd.push(Buffer.from(cfg.InternalNetworkId).toString());
        }
    });
    UpdatePlcManagerLinks(kscApi, hostId, changedData.UsedInternalIds, InternalIdsOnEnd);
}

/*
function RbacToPolicyCache(activePolicyCache){
    return activePolicyCache;
}
*/

module.exports = {
    defPolicySections,
    toView,
    toRawEx,
//    RbacToPolicyCache
};
