const { helpers } = require('@kl/openapi-client')

const lo = require('lodash')
const fs = require('fs')
const path = require('path')

const consts = require('./consts')
const hlp = require('./helpers')

function RemoveMandatories(item) {
    Object.keys(item).forEach(property => {
        if(item[property].KLPRSS_Mnd!==undefined){
            item[property].KLPRSS_Mnd = false;
        }
        if(item[property].KLPRSS_Val!== undefined){
            if(typeof item[property].KLPRSS_Val[0]==='object' && !Array.isArray(item[property].KLPRSS_Val[0]) ){
                RemoveMandatories(item[property].KLPRSS_Val[0]);
            }
            if(typeof item[property].KLPRSS_Val==='object' && item[property].KLPRSS_Val.type===undefined ){
                RemoveMandatories(item[property].KLPRSS_Val);
            }
        }
    })
}

function DefaultPolicyObject(policySection, sectionMapper){
    try {
        const policyShortName = sectionMapper!==undefined ? sectionMapper[policySection] : policySection;
        const parsedFileName = path.parse(policyShortName + '.json');
        const filePath = path.resolve(__dirname, 'converters', 'policy', 'defaults', parsedFileName.base);
        runtime.logger.log(consts.logProductPref + ' DefaultPolicyObject: Read file: ' + filePath);
        const DefaultsRaw = fs.readFileSync(filePath, 'utf-8');
        return hlp.deserialize({PxgRetVal:JSON.parse(DefaultsRaw)}).KLPRSS_Val;
    } catch (error) {
        console.log(consts.logProductPref + ' Failed to read default section ' + policySection + '.json');
    }
    return {};
}

async function GetHostActivePolicyId(kscApi, hostId) {
    let policy = await kscApi.utils.getHostActivePolicy({
        hostName:hostId,
        productId: consts.wstrProduct,
        productVersion: consts.wstrVersion,
    });
    return (policy!==undefined && policy.KLPOL_ID!==undefined ) ? policy.KLPOL_ID : -1
}

function UpdatePoliciesCacheAsDefault(arrSections, policyCache, sectionMapper) {
    arrSections = Array.isArray(arrSections) ? arrSections : [arrSections]
    for (let section of arrSections) {
        if( section==='' )
            continue;
        const rowPolicy = DefaultPolicyObject(section, sectionMapper);
        let defPolicy = rowPolicy.Policy!==undefined ? rowPolicy.Policy.KLPRSS_Val : rowPolicy;
        RemoveMandatories(defPolicy);
        policyCache[section] = defPolicy;
    }
}

async function UpdatePoliciesCache(kscApi, policyId, arrSections, policyCache) {
    if (policyId === -1 || policyId===undefined)
        return;

    arrSections = Array.isArray(arrSections) ? arrSections : [arrSections]

    let { result: identifier } = await kscApi.Policy.GetPolicyContents({nPolicy:policyId});
    for (let section of arrSections) {
        if( consts.policySsSections[section]===undefined )
            continue;

        let sectionData = await kscApi.SsContents.Ss_Read({
            wstrID: identifier,
            wstrProduct: consts.wstrProduct,
            wstrVersion: consts.wstrVersion,
            wstrSection: consts.policySsSections[section],
            skipDeserialization: true});

        if(sectionData.result!==undefined){
            const rowData = hlp.deserialize(lo.get(sectionData, 'result')).KLPRSS_Val;
            policyCache[section] = rowData.Policy.KLPRSS_Val;
        } else {
            policyCache[section] = DefaultPolicyObject(section);
        }
        runtime.logger.log(consts.logProductPref + ' UpdatePoliciesCache: ' + section + ' Data:' + JSON.stringify(policyCache[section]));
    }
    kscApi.SsContents.Ss_Release({ wstrID: identifier });
}

async function UpdatePoliciesCacheEx(kscApi, policyId, arrSections, policyCache, sectionMapper) {
    if (policyId === -1 || policyId===undefined)
        return;

    arrSections = Array.isArray(arrSections) ? arrSections : [arrSections];

    let { result: identifier } = await kscApi.Policy.GetPolicyContents({nPolicy:policyId});
    for (let section of arrSections) {
        const section_name = sectionMapper && sectionMapper[section] ? sectionMapper[section] : section;
        if( consts.policySsSections[section_name]===undefined )
            continue;

        let sectionData = await kscApi.SsContents.Ss_Read({
            wstrID: identifier,
            wstrProduct: consts.wstrProduct,
            wstrVersion: consts.wstrVersion,
            wstrSection: consts.policySsSections[section_name],
            skipDeserialization: true});

        if(sectionData.result!==undefined){
            const rowData = hlp.deserialize(lo.get(sectionData, 'result')).KLPRSS_Val;
            policyCache[section] = rowData.Policy.KLPRSS_Val;
        } else {
            policyCache[section] = DefaultPolicyObject(section);
        }
    }
    kscApi.SsContents.Ss_Release({ wstrID: identifier });
}

function SetPolicyFieldsRbacLocked(changingData, fields) {
    fields = (fields!==undefined && Array.isArray(fields)) ? fields : [];
    fields.forEach(function (item) {
        if (item !== '__VersionInfo' && changingData[item] !== undefined) {
            changingData[item].KLPRSS_Mnd = true;
            changingData[item].KLPRSS_ValLck = true;
            changingData[item].KLPRSS_ValLckGroup = 'Rbac';
        }
    });
    return changingData;
}

async function GetTaskData(kscApi, taskId, hostId) {
    let result = {}
    const strTaskId = taskId.toString();
    try {
        let task = {}
        if (hostId){
            const strHostName = hostId.toString()
            let {result: strSrvObjId} = await kscApi.HostGroup.GetHostTasks({strHostName})
            task = await kscApi.HostTasks.GetTaskData({strTask:strTaskId, strSrvObjId})
            result = task.result
        }
        else {
            task = await kscApi.Tasks.GetTaskData({taskId:strTaskId, skipDeserialization:true})
            result = helpers.serializer.deserializeKscRequest({body: task.result}).result
        }
    } catch (e) {
        console.error(e);
    }
    return result
}

module.exports = {
    DefaultPolicyObject,
    GetHostActivePolicyId,
    UpdatePoliciesCache,
    UpdatePoliciesCacheEx,
    UpdatePoliciesCacheAsDefault,
    SetPolicyFieldsRbacLocked,
    //
    GetTaskData
}